from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Callable, Self, overload

import dateparser  # type: ignore

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.autodoc.schema import Datetime as AutoDocDatetime
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.column import Column
from clearskies.query import Condition

if TYPE_CHECKING:
    from clearskies import Model


class Datetime(Column):
    """
    Stores date+time data in a column.

    When processing user input, this value is passed through `dateparser.parse()` to decide if it is a proper date string.
    This makes for relatively flexible input validation.  Example:

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        my_datetime = clearskies.columns.Datetime()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyModel,
            writeable_column_names=["name", "my_datetime"],
            readable_column_names=["id", "name", "my_datetime"],
        ),
        classes=[MyModel],
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080' -d '{"name":"Bob", "my_datetime":"2025-05-13 12:35:45+00:00"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "68095d0d-c909-4ab3-8c15-bd2667b7b074",
            "name": "Bob",
            "my_datetime": "2025-05-13T12:35:45+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080' -d '{"name":"Bob", "my_datetime":"May 13th 2025 2:35:45UTC"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "9fea6933-86ac-4dd1-b9e0-a9fa50608410",
            "name": "Bob",
            "my_datetime": "2025-05-13T12:35:45+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080' -d '{"name":"Bob", "my_datetime":"not a date"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "my_datetime": "given value did not appear to be a valid date"
        }
    }
    ```
    """

    """
    Whether or not to make datetime objects timezone-aware
    """
    timezone_aware = configs.Boolean(default=True)

    """
    The timezone to use for the datetime object (if it is timezone aware)
    """
    timezone = configs.Timezone(default=datetime.timezone.utc)

    """
    The format string to use when sending to the backend (default: %Y-%m-%d %H:%M:%S)
    """
    date_format = configs.String(default="%Y-%m-%d %H:%M:%S")

    """
    A default value to set for this column.

    The default is only used when creating a record for the first time, and only if
    a value for this column has not been set.
    """
    default = configs.Datetime()  # type: ignore

    """
    Sets a default date that the backend is going to provide.

    Some backends, depending on configuration, may provide a default value for the column
    instead of null.  By setting this equal to that default value, clearskies can detect
    when a given value is actually a non-value.
    """
    backend_default = configs.String(default="0000-00-00 00:00:00")

    setable = configs.DatetimeOrCallable(default=None)  # type: ignore
    _allowed_search_operators = ["<=>", "!=", "<=", ">=", ">", "<", "=", "in", "is not null", "is null"]
    auto_doc_class: type[AutoDocSchema] = AutoDocDatetime
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        date_format: str = "%Y-%m-%d %H:%M:%S",
        backend_default: str = "0000-00-00 00:00:00",
        timezone_aware: bool = True,
        timezone: datetime.timezone = datetime.timezone.utc,
        default: datetime.datetime | None = None,
        setable: datetime.datetime | Callable[..., datetime.datetime] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def from_backend(self, value) -> datetime.datetime | None:
        if not value or value == self.backend_default:
            return None
        if isinstance(value, str):
            value = dateparser.parse(value)
        if not isinstance(value, datetime.datetime):
            raise TypeError(
                f"I was expecting to get a datetime from the backend but I didn't get anything recognizable.  I have a value of type '{value.__class__.__name__}'.  I need either a datetime object or a datetime serialized as a string."
            )
        if self.timezone_aware:
            if not value.tzinfo:
                value = value.replace(tzinfo=self.timezone)
            elif value.tzinfo != self.timezone:
                value = value.astimezone(self.timezone)
        else:
            value = value.replace(tzinfo=None)

        return value

    def to_backend(self, data: dict[str, Any]) -> dict[str, Any]:
        if self.name not in data or isinstance(data[self.name], str) or data[self.name] is None:
            return data

        value = data[self.name]
        if not isinstance(data[self.name], datetime.datetime):
            raise TypeError(
                f"I was expecting a stringified-date or a datetime object to send to the backend, but instead I found a value of {value.__class__.__name__}"
            )

        return {
            **data,
            self.name: value.strftime(self.date_format),
        }

    def to_json(self, model: clearskies.model.Model) -> dict[str, Any]:
        """Grabs the column out of the model and converts it into a representation that can be turned into JSON."""
        value = self.__get__(model, model.__class__)
        if value and (isinstance(value, datetime.datetime) or isinstance(value, datetime.date)):
            value = value.isoformat()  # type: ignore

        return {self.name: value}

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> datetime.datetime:
        pass

    def __get__(self, instance, cls):
        return super().__get__(instance, cls)

    def __set__(self, instance, value: datetime.datetime) -> None:
        instance._next_data[self.name] = value

    def equals(self, value: str | datetime.datetime) -> Condition:
        return super().equals(value)

    def spaceship(self, value: str | datetime.datetime) -> Condition:
        return super().spaceship(value)

    def not_equals(self, value: str | datetime.datetime) -> Condition:
        return super().not_equals(value)

    def less_than_equals(self, value: str | datetime.datetime) -> Condition:
        return super().less_than_equals(value)

    def greater_than_equals(self, value: str | datetime.datetime) -> Condition:
        return super().greater_than_equals(value)

    def less_than(self, value: str | datetime.datetime) -> Condition:
        return super().less_than(value)

    def greater_than(self, value: str | datetime.datetime) -> Condition:
        return super().greater_than(value)

    def is_in(self, values: list[str | datetime.datetime]) -> Condition:
        return super().is_in(values)

    def input_error_for_value(self, value, operator=None):
        value = dateparser.parse(value)
        if not value:
            return "given value did not appear to be a valid date"
        if not value.tzinfo and self.timezone_aware:
            return "date is missing timezone information"
        return ""

    def values_match(self, value_1, value_2):
        """Compare two values to see if they are the same."""
        # in this function we deal with data directly out of the backend, so our date is likely
        # to be string-ified and we want to look for default (e.g. null) values in string form.
        if type(value_1) == str and ("0000-00-00" in value_1 or value_1 == self.backend_default):
            value_1 = None
        if type(value_2) == str and ("0000-00-00" in value_2 or value_2 == self.backend_default):
            value_2 = None
        number_values = 0
        if value_1:
            number_values += 1
        if value_2:
            number_values += 1
        if number_values == 0:
            return True
        if number_values == 1:
            return False

        if type(value_1) == str:
            value_1 = dateparser.parse(value_1)
        if type(value_2) == str:
            value_2 = dateparser.parse(value_2)

        # we need to make sure we're comparing in the same timezones.  For our purposes, a difference in timezone
        # is fine as long as they represent the same time (e.g. 16:00EST == 20:00UTC).  For python, same time in different
        # timezones is treated as different datetime objects.
        if value_1.tzinfo is not None and value_2.tzinfo is not None:
            value_1 = value_1.astimezone(value_2.tzinfo)

        # two times can be the same but if one is datetime-aware and one is not, python will treat them as not equal.
        # we want to treat such times as being the same.  Therefore, check for equality but ignore the timezone.
        for to_check in ["year", "month", "day", "hour", "minute", "second", "microsecond"]:
            if getattr(value_1, to_check) != getattr(value_2, to_check):
                return False

        # and since we already converted the timezones to match (or one has a timezone and one doesn't), we're good to go.
        # if we passed the above loop then the times are the same.
        return True



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import clearskies.configs
from clearskies import parameters_to_properties
from clearskies.validator import Validator

if TYPE_CHECKING:
    import clearskies.model


class Required(Validator):
    is_required = True

    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        # you'd think that "required" is straight forward and we want an input error if it isn't found.
        # this isn't strictly true though.  If the model already exists, the column has a value in the model already,
        # and the column is completely missing from the input data, then it is actually perfectly fine (because
        # there will still be a value in the column after the save).  However, if the model doesn't exist, then
        # we must require the column in the data with an actual value.
        has_value = False
        has_some_value = False
        if column_name in data:
            has_some_value = True
            if type(data[column_name]) == str:
                has_value = bool(data[column_name].strip())
            else:
                has_value = bool(data[column_name])
        if has_value:
            return ""
        if model and getattr(model, column_name) and not has_some_value:
            return ""
        return f"'{column_name}' is required."



from clearskies.configs import config


class StringList(config.Config):
    """
    This is for a configuration that should be a list of strings.

    This is different than SelectList, which also accepts a list of strings, but
    valdiates that all of those values match against an allow list.
    """

    def __set__(self, instance, value: list[str]):
        if value is None:
            return

        if not isinstance(value, list):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a list parameter"
            )
        for index, item in enumerate(value):
            if not isinstance(item, str):
                error_prefix = self._error_prefix(instance)
                raise TypeError(
                    f"{error_prefix} attempt to set a value of type '{item.__class__.__name__}' for item #{index + 1}.  A string was expected."
                )
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> list[str]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from clearskies.configs import string
from clearskies.functional import routing


class Url(string.String):
    def __set__(self, instance, value: str):
        if value is None:
            return

        if not isinstance(value, str):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a url parameter"
            )
        value = value.strip("/")

        if value:
            try:
                routing.extract_url_parameter_name_map(value)
            except ValueError as e:
                error_prefix = self._error_prefix(instance)
                raise ValueError(f"{error_prefix} {e}")
        instance._set_config(self, value)



import json
from typing import Any

import clearskies.configs
import clearskies.di
import clearskies.parameters_to_properties
from clearskies.authentication.authentication import Authentication
from clearskies.exceptions import ClientError
from clearskies.security_headers.cors import Cors


class Jwks(Authentication, clearskies.di.InjectableProperties):
    """The URL where the JWKS can be found."""

    jwks_url = clearskies.configs.String(required=True)

    """
    The audience to accept JWTs for.
    """
    audience = clearskies.configs.StringList(default=[])

    """
    The expected issuer of the JWTs.
    """
    issuer = clearskies.configs.String(default="")

    """
    The allowed algorithms
    """
    algorithms = clearskies.configs.StringList(default=["RS256"])

    """
    The number of seconds for which the JWKS URL contents can be cached
    """
    jwks_cache_time = clearskies.configs.Integer(default=86400)

    """
    The Authorization URL (used in the auto-generated documentation)
    """
    authorization_url = clearskies.configs.String()

    """
    The name of the security scheme in the auto-generated documentation.
    """
    documentation_security_name = clearskies.configs.String(default="jwt")

    """
    The environment helper.
    """
    environment = clearskies.di.inject.Environment()

    """
    The requests object.
    """
    requests = clearskies.di.inject.Requests()

    """
    The JoseJwt library
    """
    jose_jwt = clearskies.di.inject.ByName("jose_jwt")

    """
    The current time
    """
    now = clearskies.di.inject.Now()

    """
    Local cache of the JWKS
    """
    _jwks = None

    """
    The time when the JWKS was last fetched
    """
    _jwks_fetched = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        jwks_url: str,
        audience: str = "",
        issuer: str = "",
        algorithms: list[str] = ["RS256"],
        jwks_cache_time: int = 86400,
        authorization_url: str = "",
        documentation_security_name: str = "jwt",
    ):
        self.finalize_and_validate_configuration()

    def authenticate(self, input_output) -> bool:
        auth_header = input_output.get_request_header("authorization", True)
        if not auth_header:
            raise ClientError("Missing 'Authorization' header in request")
        if auth_header[:7].lower() != "bearer ":
            raise ClientError("Missing 'Bearer ' prefix in authorization header")
        self.validate_jwt(auth_header[7:])
        input_output.authorization_data = self.jwt_claims
        return True

    def validate_jwt(self, raw_jwt):
        try:
            from jwcrypto import jwk, jws, jwt  # type: ignore
            from jwcrypto.common import JWException  # type: ignore
        except:
            raise ValueError(
                "The JWKS authentication method requires the jwcrypto libraries to be installed.  These are optional dependencies of clearskies, so to include them do a `pip install 'clear-skies[jwcrypto]'`"
            )

        keys = jwk.JWKSet()
        keys.import_keyset(json.dumps(self._get_jwks()))

        client_jwt = jwt.JWT()
        try:
            client_jwt.deserialize(raw_jwt)
        except Exception as e:
            raise ClientError(str(e))

        try:
            client_jwt.validate(keys)
            self.jwt_claims = json.loads(client_jwt.claims)
        except JWException as e:
            raise ClientError(str(e))

        if self.issuer and self.jwt_claims.get("iss") != self.issuer:
            raise ClientError("Issuer does not match")

        if self.audience:
            jwt_audience = self.jwt_claims.get("aud")
            if not jwt_audience:
                raise ClientError("Audience required, but missing in JWT")
            has_match = False
            for audience in jwt_audience:
                if audience == self.audience:
                    has_match = True
            if not has_match:
                raise ClientError("Audience does not match")

        return True

    def _get_jwks(self):
        if self._jwks is None or ((self.now - self._jwks_fetched).total_seconds() > self.jwks_cache_time):
            self._jwks = self.requests.get(self.jwks_url).json()
            self._jwks_fetched = self.now

        return self._jwks

    def documentation_security_scheme(self) -> dict[str, Any]:
        return {
            "type": "oauth2",
            "description": "JWT based authentication",
            "flows": {"implicit": {"authorizationUrl": self.authorization_url, "scopes": {}}},
        }

    def documentation_security_scheme_name(self) -> str:
        return self.documentation_security_name

    def set_headers_for_cors(self, cors: Cors):
        cors.add_header("Authorization")



from __future__ import annotations

from typing import TYPE_CHECKING, Self, Type, overload

from clearskies.columns import CategoryTreeChildren

if TYPE_CHECKING:
    from clearskies import Model


class CategoryTreeDescendants(CategoryTreeChildren):
    """
    Return all descendants from a category tree column.

    See the CategoryTree column for usage examples.

    The descendants are the recursive children of a given category.  So, given the following tree:

    ```
    Root/
    ├─ Sub/
    │  ├─ Sub Sub/
    │  │  ├─ Sub Sub Sub/
    ├─ Another Child/

    The descendants of `Root` are `["Sub", "Sub Sub", "Sub Sub Sub", "Another Child"]`.
    """

    _descriptor_config_map = None

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> Model:
        pass

    def __get__(self, model, cls):
        if model is None:
            self.model_class = cls
            return self  # type:  ignore

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            model.get_columns()

        return self.relatives(model, include_all=True)



from abc import ABC, abstractmethod
from typing import Any


class Injectable(ABC):
    _di: Any = None

    def initiated_guard(self, instance):
        if self._di:
            return

        reference = instance.__class__.__name__ + "."
        my_id = id(self)
        cls = instance.__class__
        for attribute_name in dir(instance):
            if id(getattr(cls, attribute_name)) != my_id:
                continue
            reference += attribute_name
        raise ValueError(
            f"There was an attempt to get a value out of '{reference}' but the injectable hasn't been properly"
            + "initialized.  This usually means that objects are being created outside of the normal Di system."
        )

    def set_di(self, di):
        self._di = di

    @abstractmethod
    def __get__(self, instance, parent):
        pass



from clearskies.query.condition import Condition, ParsedCondition
from clearskies.query.join import Join
from clearskies.query.query import Query
from clearskies.query.sort import Sort

__all__ = [
    "Condition",
    "Join",
    "ParsedCondition",
    "Sort",
    "Query",
]



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from clearskies import Column


class Schema:
    """
    Define a schema by extending and declaring columns.

    ```python
    from clearskies.schema import Schema
    from clearskies.validators import Required, Unique

    import clearskies.columns


    class Person(Schema):
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String(validators=[Required()])
        date_of_birth = clearskies.columns.Datetime(validators=[Required(), InThePast()])
        email = clearskies.columns.Email()
    ```
    """

    id_column_name: str = ""
    _columns: dict[str, Column] = {}

    @classmethod
    def destination_name(cls: type[Self]) -> str:
        raise NotImplementedError()

    def __init__(self):
        self._data = {}

    @classmethod
    def get_columns(cls: type[Self], overrides={}) -> dict[str, Column]:
        """
        Return an ordered dictionary with the configuration for the columns.

        Generally, this method is meant for internal use.  It just pulls the column configuration
        information out of class attributes.  It doesn't return the fully prepared columns,
        so you probably can't use the return value of this function.  For that, see
        `model.columns()`.
        """
        # no caching if we have overrides
        if cls._columns and not overrides:
            return cls._columns

        overrides = {**overrides}
        columns: dict[str, Column] = OrderedDict()
        for attribute_name in dir(cls):
            attribute = getattr(cls, attribute_name)
            # use duck typing instead of isinstance to decide which attribute is a column.
            # We have to do this to avoid circular imports.
            if not hasattr(attribute, "from_backend") and not hasattr(attribute, "to_backend"):
                continue

            if attribute_name in overrides:
                columns[attribute_name] = overrides[attribute_name]
                del overrides[attribute_name]
            columns[attribute_name] = attribute

        for attribute_name, column in overrides.items():
            columns[attribute_name] = column  # type: ignore

        if not overrides:
            cls._columns = columns

        # now go through and finalize everything.  We have to do this after setting cls._columns, because finalization
        # sometimes depends on fetching the list of columns, so if we do it before caching the answer, we may end up
        # creating circular loops.  I don't *think* this will cause painful side-effects, but we'll find out!
        for column_name, column in cls._columns.items():
            column.finalize_configuration(cls, column_name)

        return columns

    def __bool__(self):
        return False



class Schema:
    name = None
    example = None
    value = None
    _type = "string"
    _format = ""

    def __init__(self, name, example=None, value=None):
        self.name = name
        self.example = example
        self.value = value



from __future__ import annotations

from typing import TYPE_CHECKING

import requests

from clearskies.di.injectable import Injectable

if TYPE_CHECKING:
    from clearskies.input_outputs.input_output import InputOutput as InputOuputDependency


class InputOutput(Injectable):
    def __init__(self):
        pass

    def __get__(self, instance, parent) -> InputOuputDependency:
        if instance is None:
            return self  # type: ignore
        return self._di.build_from_name("input_output", cache=True)



from clearskies.configurable import Configurable


class SecurityHeader(Configurable):
    is_cors = False

    def set_headers_for_input_output(self, input_output):
        raise NotImplementedError()



from clearskies.configs import config


class StringDict(config.Config):
    """This is for a configuration that should be a dictionary with keys and values that are all strings."""

    def __set__(self, instance, value: dict[str, str]):
        if value is None:
            return

        if not isinstance(value, dict):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a dict parameter"
            )
        for key, val in value.items():
            if not isinstance(key, str):
                error_prefix = self._error_prefix(instance)
                raise TypeError(
                    f"{error_prefix} attempt to set a key of type '{key.__class__.__name__}' when only string keys are allowed."
                )
            if not isinstance(val, str):
                error_prefix = self._error_prefix(instance)
                raise TypeError(
                    f"{error_prefix} attempt to set a value of type '{val.__class__.__name__}' for key '{key}'.  A string was expected."
                )
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> dict[str, str]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from .schema import Schema


class Enum(Schema):
    values = None
    value_type = None
    example = None
    _type = "string"
    _format = ""

    def __init__(self, name, values, value_type, example=None):
        self.name = name
        self.values = values
        self.value_type = value_type
        self._type = value_type._type
        self._format = value_type._format
        self.example = example



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Self, overload

import clearskies.configs.actions
import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.autodoc.schema import Boolean as AutoDocBoolean
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.column import Column
from clearskies.query import Condition

if TYPE_CHECKING:
    from clearskies import Model


class Boolean(Column):
    """Represents a column with a true/false type."""

    """
    Actions to trigger when the column changes to True
    """
    on_true = clearskies.configs.actions.Actions(default=[])

    """
    Actions to trigger when the column changes to False
    """
    on_false = clearskies.configs.actions.Actions(default=[])

    """
    The class to use when documenting this column
    """
    auto_doc_class: type[AutoDocSchema] = AutoDocBoolean

    _allowed_search_operators = ["="]
    default = configs.Boolean()  #  type: ignore
    setable = configs.BooleanOrCallable()  #  type: ignore
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        default: bool | None = None,
        setable: bool | Callable[..., bool] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_true: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_false: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def from_backend(self, value) -> bool:
        if value == "0":
            return False
        return bool(value)

    def to_backend(self, data):
        if self.name not in data:
            return data

        return {**data, self.name: bool(data[self.name])}

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> bool:
        pass

    def __get__(self, instance, cls):
        return super().__get__(instance, cls)

    def __set__(self, instance, value: bool) -> None:
        instance._next_data[self.name] = value

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        return f"{self.name} must be a boolean" if type(value) != bool else ""

    def build_condition(self, value: str, operator: str | None = None, column_prefix: str = ""):
        condition_value = "1" if value else "0"
        if not operator:
            operator = "="
        return f"{column_prefix}{self.name}{operator}{condition_value}"

    def save_finished(self, model: Model) -> None:
        """Make any necessary changes needed after a save has completely finished."""
        super().save_finished(model)

        if (not self.on_true and not self.on_false) or not model.was_changed(self.name):
            return

        if getattr(model, self.name) and self.on_true:
            self.execute_actions(self.on_true, model)
        if not getattr(model, self.name) and self.on_false:
            self.execute_actions(self.on_false, model)

    def equals(self, value: bool) -> Condition:
        return super().equals(value)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import clearskies.di
import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.columns.string import String

if TYPE_CHECKING:
    from clearskies import Model


class Uuid(String):
    """
    Populates the column with a UUID upon record creation.

    This column really just has a very specific purpose: ids!

    When used, it will automatically populate the column with a random UUID upon record creation.
    It is not a writeable column, which means that you cannot expose it for write operations via an endpoint.

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyModel,
            writeable_column_names=["name"],
            readable_column_names=["id", "name"],
        ),
    )
    wsgi()
    ```

    and when invoked:

    ```bash
    $ curl http://localhost:8080 -d '{"name": "John Doe"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "d4f23106-b48a-4dc5-9bf6-df61f6ca54f7",
            "name": "John Doe"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```
    """

    is_writeable = configs.Boolean(default=False)
    _descriptor_config_map = None

    uuid = clearskies.di.inject.Uuid()

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        is_readable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        pass

    def pre_save(self, data: dict[str, Any], model: Model) -> dict[str, Any]:
        if model:
            return data
        data = {**data, self.name: str(self.uuid.uuid4())}
        if self.on_change_pre_save:
            data = self.execute_actions_with_data(self.on_change_pre_save, model, data)
        return data



class Array:
    def __init__(self, schema, oai3_schema_resolver):
        self.schema = schema
        self.oai3_schema_resolver = oai3_schema_resolver

    def convert(self):
        schema = {"type": "array", "items": self.oai3_schema_resolver(self.schema.item_definition).convert()}

        return schema



import re

from clearskies.columns.string import String


class Email(String):
    """
    A string column that specifically expects an email.

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        email = clearskies.columns.Email()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyModel,
            writeable_column_names=["email"],
            readable_column_names=["id", "email"],
        ),
        classes=[MyModel],
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080' -d '{"email":"test@example.com"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "2a72a895-c469-45b0-b5cd-5a3cbb3a6e99",
            "email": "test@example.com"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080' -d '{"email":"asdf"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "email": "Invalid email address"
        }
    }
    ```
    """

    _descriptor_config_map = None

    """
    A column that always requires an email address.
    """

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        if type(value) != str:
            return f"Value must be a string for {self.name}"
        if operator and operator.lower() == "like":
            # don't check for an email if doing a fuzzy search, since we may be searching
            # for a partial email
            return ""
        if re.search(r"^[^@\s]+@[^@]+\.[^@]+$", value):
            return ""
        return "Invalid email address"



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable, Type

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.authentication import Authentication, Authorization
from clearskies.endpoint import Endpoint
from clearskies.functional import routing, string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import Column, Schema, SecurityHeader
    from clearskies.model import Model


class Get(Endpoint):
    """
    An endpoint that fetches a single record by id (or other unique column).

    This endpoint is intended to return a single record.  You have to provide a model class, the list of columns
    to return, and (optionally) the name of the column that it should use to lookup the matching record.  If you don't
    specifically tell it what column to use to lookup the record, it will assume that you want to use the id column
    of the model.  Finally, you must declare a route parameter with a matching column name: the get endpoint will then
    fetch the desired record id out of the URL path.  Here's a simple example:

    ```python
    import clearskies


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        username = clearskies.columns.String()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Get(
            model_class=User,
            url="/{id}",
            readable_column_names=["id", "name", "username"],
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": User,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob Brown", "username": "bobbrown"},
                        {"id": "1-2-3-5", "name": "Jane Doe", "username": "janedoe"},
                        {"id": "1-2-3-6", "name": "Greg", "username": "greg"},
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080/1-2-3-4' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "1-2-3-4",
            "name": "Bob Brown",
            "username": "bobbrown"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/1-2-3-5' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "1-2-3-5",
            "name": "Jane Doe",
            "username": "janedoe"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/notauser' | jq
    {
        "status": "client_error",
        "error": "Not Found",
        "data": [],
        "pagination": {},
        "input_errors": {}
    }
    ```
    """

    """
    Specify the name of the column that should be used to look up the record.

    If not specified, it will default to the id column name.  There must be a matching route parameter in the URL.

    ```python
    import clearskies

    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        username = clearskies.columns.String()

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Get(
            model_class=User,
            url="/{username}",
            readable_column_names=["id", "name", "username"],
            record_lookup_column_name="username",
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": User,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob Brown", "username": "bobbrown"},
                        {"id": "1-2-3-5", "name": "Jane Doe", "username": "janedoe"},
                        {"id": "1-2-3-6", "name": "Greg", "username": "greg"},
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    Note that `record_lookup_column_name` is set to `username` and we similarly changed the route from
    `/{id}` to `/{username}`.  We then invoke it with the username rather than the id:

    ```bash
    $ curl 'http://localhost:8080/janedoe' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "1-2-3-5",
            "name": "Jane Doe",
            "username": "janedoe"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```
    """
    record_lookup_column_name = clearskies.configs.ReadableModelColumn("model_class", default=None)

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        model_class: type[Model],
        url: str,
        readable_column_names: list[str],
        record_lookup_column_name: str | None = None,
        request_methods: list[str] = ["GET"],
        response_headers: list[str | Callable[..., list[str]]] = [],
        output_map: Callable[..., dict[str, Any]] | None = None,
        output_schema: Schema | None = None,
        column_overrides: dict[str, Column] = {},
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        where: typing.condition | list[typing.condition] = [],
        joins: typing.join | list[typing.join] = [],
        authentication: Authentication = authentication.Public(),
        authorization: Authorization = authentication.Authorization(),
    ):
        try:
            # we will set the value for this if it isn't already set, and the easiest way is to just fetch it and see if it blows up
            self.record_lookup_column_name
        except:
            self.record_lookup_column_name = self.model_class.id_column_name

        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__()

        route_parameters = routing.extract_url_parameter_name_map(url)
        if self.record_lookup_column_name not in route_parameters:
            raise KeyError(
                f"Configuration error for {self.__class__.__name__} endpoint: record_lookup_column_name is set to '{self.record_lookup_column_name}' but no matching routing parameter is found"
            )

    def get_model_id(self, input_output: InputOutput) -> str:
        routing_data = input_output.routing_data
        if self.record_lookup_column_name in routing_data:
            return routing_data[self.record_lookup_column_name]
        raise KeyError(
            f"I didn't receive the ID in my routing data.  I am probably misconfigured.  My record_lookup_column_name is '{self.record_lookup_column_name}' and my route is {self.url}"
        )

    def fetch_model(self, input_output: InputOutput) -> Model:
        lookup_column_value = self.get_model_id(input_output)
        model = self.fetch_model_with_base_query(input_output).find(
            self.record_lookup_column_name + "=" + lookup_column_value
        )
        if not model:
            raise clearskies.exceptions.NotFound("Not Found")
        return model

    def handle(self, input_output: InputOutput) -> Any:
        model = self.fetch_model(input_output)
        return self.success(input_output, self.model_as_json(model, input_output))

    def documentation(self) -> list[autodoc.request.Request]:
        output_schema = self.model_class
        nice_model = string.camel_case_to_words(output_schema.__name__)

        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)
        output_data_schema = self.documentation_data_schema(output_schema, self.readable_column_names)
        output_autodoc = (
            autodoc.schema.Object(
                self.auto_case_internal_column_name("data"), children=output_data_schema, model_name=schema_model_name
            ),
        )

        authentication = self.authentication
        standard_error_responses = [self.documentation_input_error_response()]
        if not getattr(authentication, "is_public", False):
            standard_error_responses.append(self.documentation_access_denied_response())
            if getattr(authentication, "can_authorize", False):
                standard_error_responses.append(self.documentation_unauthorized_response())

        return [
            autodoc.request.Request(
                self.description,
                [
                    self.documentation_success_response(
                        output_autodoc,  # type: ignore
                        description=self.description,
                    ),
                    *standard_error_responses,
                    self.documentation_generic_error_response(),
                ],
                relative_path=self.url,
                request_methods=self.request_methods,
                parameters=[
                    *self.documentation_routing_parameters(),
                    *self.standard_url_parameters(),
                ],
                root_properties={
                    "security": self.documentation_request_security(),
                },
            ),
        ]

    def documentation_routing_parameters(self) -> list[autodoc.request.Parameter]:
        return self.standard_url_request_parameters()

    def documentation_models(self) -> dict[str, autodoc.schema.Schema]:
        output_schema = self.output_schema if self.output_schema else self.model_class
        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)

        return {
            schema_model_name: autodoc.schema.Object(
                self.auto_case_internal_column_name("data"),
                children=self.documentation_data_schema(output_schema, self.readable_column_names),
            ),
        }



from clearskies.configs import config


class Float(config.Config):
    def __set__(self, instance, value: float):
        if not isinstance(value, float):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to parameter that requires a float."
            )
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> float:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

from .condition import Condition
from .join import Join
from .sort import Sort

if TYPE_CHECKING:
    from clearskies import Model


class Query:
    """
    Track the various aspects of a query.

    This is mostly just used by the Model class to keep track of a list request.
    """

    """
    The model class
    """
    model_class: type[Model] = None  # type: ignore

    """
    The list of where conditions for the query.
    """
    conditions: list[Condition] = []

    """
    The conditions, but organized by column.
    """
    conditions_by_column: dict[str, list[Condition]] = {}

    """
    Joins for the query.
    """
    joins: list[Join] = []

    """
    The sort directives for the query
    """
    sorts: list[Sort] = []

    """
    The maximum number of records to return.
    """
    limit: int = 0

    """
    Pagination information (e.g. start/next_token/etc... the details depend on the backend.
    """
    pagination: dict[str, Any] = {}

    """
    A list of select statements.
    """
    selects: list[str] = []

    """
    Whether or not to just select all columns.
    """
    select_all: bool = True

    """
    The name of the column to group by.
    """
    group_by = ""

    def __init__(
        self,
        model_class: type[Model],
        conditions: list[Condition] = [],
        joins: list[Join] = [],
        sorts: list[Sort] = [],
        limit: int = 0,
        group_by: str = "",
        pagination: dict[str, Any] = {},
        selects: list[str] = [],
        select_all: bool = True,
    ):
        self.model_class = model_class
        self.conditions = [*conditions]
        self.joins = [*joins]
        self.sorts = [*sorts]
        self.limit = limit
        self.group_by = group_by
        self.pagination = {**pagination}
        self.selects = [*selects]
        self.select_all = select_all
        self.conditions_by_column = {}
        if conditions:
            for condition in conditions:
                if condition.column_name not in self.conditions_by_column:
                    self.conditions_by_column[condition.column_name] = []
                self.conditions_by_column[condition.column_name].append(condition)

    def as_kwargs(self):
        """Return the properties of this query as a dictionary so it can be used as kwargs when creating another one."""
        return {
            "model_class": self.model_class,
            "conditions": self.conditions,
            "joins": self.joins,
            "sorts": self.sorts,
            "limit": self.limit,
            "group_by": self.group_by,
            "pagination": self.pagination,
            "selects": self.selects,
            "select_all": self.select_all,
        }

    def add_where(self, condition: Condition) -> Self:
        self.validate_column(condition.column_name, "filter", table=condition.table_name)
        new_kwargs = self.as_kwargs()
        new_kwargs["conditions"].append(condition)
        return self.__class__(**new_kwargs)

    def add_join(self, join: Join) -> Self:
        new_kwargs = self.as_kwargs()
        new_kwargs["joins"].append(join)
        return self.__class__(**new_kwargs)

    def set_sort(self, sort: Sort, secondary_sort: Sort | None = None) -> Self:
        self.validate_column(sort.column_name, "sort", table=sort.table_name)
        new_kwargs = self.as_kwargs()
        new_kwargs["sorts"] = [sort]
        if secondary_sort:
            new_kwargs["sorts"].append(secondary_sort)

        return self.__class__(**new_kwargs)

    def set_limit(self, limit: int) -> Self:
        if not isinstance(limit, int):
            raise TypeError(
                f"The limit in a query must be of type int but I received a value of type '{limit.__class__.__name__}'"
            )
        return self.__class__(
            **{
                **self.as_kwargs(),
                "limit": limit,
            }
        )

    def set_group_by(self, column_name) -> Self:
        self.validate_column(column_name, "group")
        return self.__class__(
            **{
                **self.as_kwargs(),
                "group_by": column_name,
            }
        )

    def set_pagination(self, pagination: dict[str, Any]) -> Self:
        return self.__class__(
            **{
                **self.as_kwargs(),
                "pagination": pagination,
            }
        )

    def add_select(self, select: str) -> Self:
        new_kwargs = self.as_kwargs()
        new_kwargs["selects"].append(select)
        return self.__class__(**new_kwargs)

    def set_select_all(self, select_all: bool) -> Self:
        return self.__class__(
            **{
                **self.as_kwargs(),
                "select_all": select_all,
            }
        )

    def validate_column(self: Self, column_name: str, action: str, table: str | None = None) -> None:
        # for now, only validate columns that belong to *our* table.
        # in some cases we are explicitly told the column name
        if table is not None:
            # note that table may be '', in which case it is implicitly "our" table
            if table != "" and table != self.model_class.destination_name():
                return

        # but in some cases we should check and see if it is included with the column name
        column_name = column_name.replace("`", "")
        if "." in column_name:
            parts = column_name.split(".")
            if parts[0] != self.model_class.destination_name():
                return
            column_name = column_name.split(".")[1]

        model_columns = self.model_class.get_columns()
        if column_name not in model_columns:
            raise KeyError(
                f"Cannot {action} by column '{column_name}' for model class {self.model_class.__name__} because this "
                + "column does not exist for the model.  You can suppress this error by adding a matching column "
                + "to your model definition"
            )



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.endpoint import Endpoint
from clearskies.functional import string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import Column, SecurityHeader
    from clearskies.model import Model


class Create(Endpoint):
    """
    An endpoint to create a record.

    This endpoint accepts user input and uses it to create a record for the given model class.  You have
    to provide the model class, which columns the end-user can set, and which columns get returned
    to the client.  The column definitions in the model class are used to strictly validate the user
    input.  Here's a basic example of a model class with the create endpoint in use:

    ```python
    import clearskies
    from clearskies import validators, columns


    class MyAwesomeModel(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = columns.Uuid()
        name = clearskies.columns.String(
            validators=[
                validators.Required(),
                validators.MaximumLength(50),
            ]
        )
        email = columns.Email(validators=[validators.Unique()])
        some_number = columns.Integer()
        expires_at = columns.Date()
        created_at = columns.Created()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyAwesomeModel,
            readable_column_names=["id", "name", "email", "some_number", "expires_at", "created_at"],
            writeable_column_names=["name", "email", "some_number", "expires_at"],
        ),
    )
    wsgi()
    ```

    The following shows how to invoke it, and demonstrates the strict input validation that happens as part of the
    process:

    ```bash
    $ curl 'http://localhost:8080/' -d '{"name":"Example", "email":"test@example.com","some_number":5,"expires_at":"2024-12-31"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "74eda1c6-fe66-44ec-9246-758d16e1a304",
            "name": "Example",
            "email": "test@example.com",
            "some_number": 5,
            "expires_at": "2024-12-31",
            "created_at": "2025-05-23T16:36:30+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/' -d '{"name":"", "email":"test@example.com","some_number":"asdf","expires_at":"not-a-date", "not_a_column": "sup"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "name": "'name' is required.",
            "email": "Invalid value for 'email': the given value already exists, and must be unique.",
            "some_number": "value should be an integer",
            "expires_at": "given value did not appear to be a valid date",
            "not_a_column": "Input column not_a_column is not an allowed input column."
        }
    }
    ```

    The first call successfully creates a new record.  The second call fails with a variety of error messages:

     1. A name wasn't provided by the model class marked this as required
     2. We provided the same email address again, but this column is marked as unique
     3. The number provided in `some_number` wasn't actually a number
     4. The provided value for `expires_at` wasn't actually a date.
     5. We provided an extra column (`not_a_column`) that wasn't in the list of allowed columns.
    """

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        model_class: type[Model],
        writeable_column_names: list[str],
        readable_column_names: list[str],
        input_validation_callable: Callable | None = None,
        include_routing_data_in_request_data: bool = False,
        url: str = "",
        request_methods: list[str] = ["POST"],
        response_headers: list[str | Callable[..., list[str]]] = [],
        output_map: Callable[..., dict[str, Any]] | None = None,
        output_schema: clearskies.Schema | None = None,
        column_overrides: dict[str, Column] = {},
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        authentication: authentication.Authentication = authentication.Public(),
        authorization: authentication.Authorization = authentication.Authorization(),
    ):
        # a bit weird, but we have to do this because the default in the above definition is different than
        # the default set on the request_mehtods config in the bsae endpoint class.  parameters_to_properties will copy
        # parameters to properties, but only for things set by the developer - not for default values set in the kwarg
        # definitions.  Therefore, we always set it here to make sure we user our default, not the one in the base class.
        self.request_methods = request_methods

        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__()

    def handle(self, input_output: InputOutput) -> Any:
        request_data = self.get_request_data(input_output)
        if not request_data and input_output.has_body():
            raise clearskies.exceptions.ClientError("Request body was not valid JSON")
        self.validate_input_against_schema(request_data, input_output, self.model_class)
        new_model = self.model.create(request_data, columns=self.columns)
        return self.success(input_output, self.model_as_json(new_model, input_output))

    def documentation(self) -> list[autodoc.request.Request]:
        output_schema = self.model_class
        nice_model = string.camel_case_to_words(output_schema.__name__)

        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)
        output_data_schema = self.documentation_data_schema(output_schema, self.readable_column_names)
        output_autodoc = (
            autodoc.schema.Object(
                self.auto_case_internal_column_name("data"), children=output_data_schema, model_name=schema_model_name
            ),
        )

        authentication = self.authentication
        standard_error_responses = [self.documentation_input_error_response()]
        if not getattr(authentication, "is_public", False):
            standard_error_responses.append(self.documentation_access_denied_response())
            if getattr(authentication, "can_authorize", False):
                standard_error_responses.append(self.documentation_unauthorized_response())

        return [
            autodoc.request.Request(
                self.description,
                [
                    self.documentation_success_response(
                        output_autodoc,  # type: ignore
                        description=self.description,
                    ),
                    *standard_error_responses,
                    self.documentation_generic_error_response(),
                ],
                relative_path=self.url,
                request_methods=self.request_methods,
                parameters=[
                    *self.documentation_request_parameters(),
                    *self.standard_url_parameters(),
                ],
                root_properties={
                    "security": self.documentation_request_security(),
                },
            ),
        ]

    def documentation_request_parameters(self) -> list[autodoc.request.Parameter]:
        return [
            *self.standard_json_request_parameters(self.model_class),
            *(self.standard_url_request_parameters() if self.include_routing_data_in_request_data else []),
        ]

    def documentation_models(self) -> dict[str, autodoc.schema.Schema]:
        output_schema = self.output_schema if self.output_schema else self.model_class
        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)

        return {
            schema_model_name: autodoc.schema.Object(
                self.auto_case_internal_column_name("data"),
                children=self.documentation_data_schema(output_schema, self.readable_column_names),
            ),
        }



from ... import schema as raw_schema
from . import schema as formatted_schema


class OAI3SchemaResolver:
    class_map = {
        raw_schema.Array: formatted_schema.Array,
        raw_schema.Enum: formatted_schema.Enum,
        raw_schema.Object: formatted_schema.Object,
    }

    def __call__(self, schema_object):
        if schema_object.__class__ in self.class_map:
            return self.class_map[schema_object.__class__](schema_object, self)
        return formatted_schema.Default(schema_object)



import inspect
from typing import Any


def is_model(to_check: Any) -> bool:
    """
    Return True/False to denote if the given value is a model instance.

    Uses ducktyping rather than checking the type, mostly to minimize the risk of circular imports
    """
    if not hasattr(to_check, "destination_name"):
        return False
    return True


def is_model_class(to_check: Any) -> bool:
    """Return True/False to denote if the given value is a model class."""
    return inspect.isclass(to_check) and is_model(to_check)


def is_model_class_reference(to_check: Any) -> bool:
    """Return True/False to denote if the given value is a reference to a model class."""
    if not hasattr(to_check, "get_model_class"):
        return False

    if inspect.isclass(to_check):
        to_check = to_check()
    return is_model_class(to_check.get_model_class())


def is_model_class_or_reference(to_check: Any, raise_error_message=False, strict=True) -> bool:
    """
    Return True/False to denote if the given value is a model class or a model reference.

    If strict is false, then it won't check that the model reference returns a model class.  This is often necessary during
    validation by configs as, otherwise, it tends to trigger circular dependency errors.
    """
    if not inspect.isclass(to_check):
        # for references we will accept either instances or classes
        if hasattr(to_check, "get_model_class"):
            return is_model_class(to_check.get_model_class()) if strict else True

        if raise_error_message:
            raise TypeError(
                f"I expected a model class or reference, but instead I received something of type '{to_check.__class__.__name__}'"
            )
        return False

    if is_model_class(to_check):
        return True

    if hasattr(to_check, "get_model_class"):
        if not strict:
            return True

        model_class = to_check().get_model_class()
        if is_model_class(model_class):
            return True
        if raise_error_message:
            raise TypeError(
                f"I expected a model class or reference.  I received a model reference of class '{to_check.__name__}', but when I fetched the model out of it, it gave me back something that wasn't a model.  Instead, I received something of type '{model_class.__name__}'"
            )

    if raise_error_message:
        raise TypeError(
            "I expected a model class or reference, but instead I received a class that was neither.  It had a type of '"
            + to_check.__class__.__name__
            + "'"
        )

    return False


def is_model_or_class(to_check: Any) -> bool:
    """Return True/False to denote if the given value is a model instance or model class."""
    return is_model(to_check) or is_model_class(to_check)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Self, Type, overload

import clearskies.configs.actions
import clearskies.configs.boolean
import clearskies.configs.select
import clearskies.configs.string
import clearskies.configs.string_or_callable
import clearskies.configs.validators
import clearskies.configurable
import clearskies.di
import clearskies.model
import clearskies.parameters_to_properties
import clearskies.typing
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.autodoc.schema import String as AutoDocString
from clearskies.query.condition import Condition, ParsedCondition
from clearskies.validator import Validator

if TYPE_CHECKING:
    from clearskies import Model, Schema


class Column(clearskies.configurable.Configurable, clearskies.di.InjectableProperties):
    """
    The base column.

    This class (well, the children that extend it) are used to define the columns that exist in a given model class -
    they help you build your schema.  The column definitions are then used by handlers and other aspects of the
    clearskies framework to automate things like input validation, front-end/backend-transformations, and more.
    Many column classes have their own configuration settings, but there are also some common configuration settings
    defined here.
    """

    """
    The column class gets the full DI container, because it does a lot of object building itself
    """
    di = clearskies.di.inject.Di()

    """
    A default value to set for this column.

    The default is only used when creating a record for the first time, and only if
    a value for this column has not been set.

    ```python
    import clearskies

    class Widget(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String(default="Jane Doe")

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            lambda widgets: widgets.create(no_data=True),
            model_class=Widget,
            readable_column_names=["id", "name"]
        ),
        classes=[Widget],
    )

    if __name__ == "__main__":
        cli()
    ```

    Which when invoked returns:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "03806afa-b189-4729-a43c-9da5aa17bf14",
            "name": "Jane Doe"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```
    """
    default = clearskies.configs.string.String(default=None)

    """
    A value to set for this column during a save operation.

    Unlike the default value, a setable value is always set during a save, even on update.  It will
    even override other values, so it is intended to be used in cases where the value is always controlled
    programmatically.

    ```python
    import clearskies
    import datetime

    class Pet(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String(setable="Spot")
        date_of_birth = clearskies.columns.Date()
        age = clearskies.columns.Integer(
            setable=lambda data, model, now:
                (now-dateparser.parse(model.latest("date_of_birth", data))).total_seconds()/(86400*365),
        )
        created = clearskies.columns.Created()

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            lambda pets: pets.create({"date_of_birth": "2020-05-03"}),
            model_class=Pet,
            readable_column_names=["id", "name", "date_of_birth", "age"]
        ),
        classes=[Pet],
    )

    if __name__ == "__main__":
        cli()
    ```

    Note the use of `model.latest()` above.  This function returns either the column information from the data array
    or, if not present, the latest column value from the model itself.  This makes it more flexible as it works
    well with both update and create operations.

    If you then execute this it will return something like:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "ec4993f4-124a-44a2-8313-816d2ad51aae",
            "name": "Spot",
            "date_of_birth": "2020-05-03",
            "age": 5,
            "created": "2025-05-03T20:23:32+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    e.g., `date_of_birth` is `age` years behind the current time (as recorded in the `created` timestamp).

    """
    setable = clearskies.configs.string_or_callable.StringOrCallable(default=None)

    """
    Whether or not this column can be converted to JSON and included in an API response.

    If this is set to False for a column and you attempt to set that column as a readable_column in an endpoint,
    clearskies will throw an exception.
    """
    is_readable = clearskies.configs.boolean.Boolean(default=True)

    """
    Whether or not this column can be set via an API call.

    If this is set to False for a column and you attempt to set the column as a writeable column in an endpoint,
    clearskies will throw an exception.
    """
    is_writeable = clearskies.configs.boolean.Boolean(default=True)

    """
    Whether or not it is possible to search by this column

    If this is set to False for a column and you attempt to set the column as a searchable column in an endpoint,
    clearskies will throw an exception.
    """
    is_searchable = clearskies.configs.boolean.Boolean(default=True)

    """
    Whether or not this column is temporary.  A temporary column is not persisted to the backend.

    Temporary columns are useful when you want the developer or end user to set a value, but you use that value to
    trigger additional behavior, rather than actually recording it.  Temporary columns often team up with actions
    or are used to calculate other values.  For instance, in our setable example above, we had both an age and
    a date of birth column, with the date of birth calculated from the age.  This obviously results in two columns
    with similar data.  One could be marked as temporary and it will be available during the save operation, but
    it will be skipped when saving data to the backend:

    ```python
    import clearskies

    class Pet(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        date_of_birth = clearskies.columns.Date(is_temporary=True)
        age = clearskies.columns.Integer(
            setable=lambda data, model, now:
                (now-dateparser.parse(model.latest("date_of_birth", data))).total_seconds()/(86400*365),
        )
        created = clearskies.columns.Created()

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            lambda pets: pets.create({"name": "Spot", "date_of_birth": "2020-05-03"}),
            model_class=Pet,
            readable_column_names=["id", "age", "date_of_birth"],
        ),
        classes=[Pet],
    )

    if __name__ == "__main__":
        cli()
    ```

    Which will return:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "ee532cfa-91cf-4747-b798-3c6dcd79326e",
            "age": 5,
            "date_of_birth": null
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    e.g. the date_of_birth column is empty.  To be clear though, it's not just empty - clearskies made no attempt to set it.
    If you were using an SQL database, you would not have to put a `date_of_birth` column in your table.

    """
    is_temporary = clearskies.configs.boolean.Boolean(default=False)

    """
    Validators to use when checking the input for this column during write operations from the API.

    Keep in mind that the validators are only checked when the column is exposed via a supporting endpoint.
    You can still set whatever values you want when saving the model directly, e.g. `model.save(...)`

    In the below example, we require a name that is at least 5 characters long, and the date of birth must
    be in the past.  Note that date of birth is not required, so the end-user can create a record without
    a date of birth.  In general, only the `Required` validator will reject a non-existent input.

    ```python
    import clearskies

    class Pet(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String(validators=[
            clearskies.validators.Required(),
            clearskies.validators.MinimumLength(5),
        ])
        date_of_birth = clearskies.columns.Date(validators=[
            clearskies.validators.InThePast()
        ])
        created = clearskies.columns.Created()

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            model_class=Pet,
            writeable_column_names=["name", "date_of_birth"],
            readable_column_names=["id", "name", "date_of_birth", "created"],
        ),
    )
    wsgi()
    ```

    You can then see the result of calling the endpoint with various kinds of invalid data:

    ```bash
    $ curl http://localhost:8080 -d '{"date_of_birth": "asdf"}'
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "name": "'name' is required.",
            "date_of_birth": "given value did not appear to be a valid date"
        }
    }

    $ curl http://localhost:8080 -d '{"name":"asdf"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "name": "'name' must be at least 5 characters long."
        }
    }

    $ curl http://localhost:8080 -d '{"name":"Longer", "date_of_birth": "2050-01-01"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "date_of_birth": "'date_of_birth' must be in the past"
        }
    }

    $ curl http://localhost:8080 -d '{"name":"Long Enough"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "ace16b93-db91-49b3-a8f7-5dc6568d25f6",
            "name": "Long Enough",
            "date_of_birth": null,
            "created": "2025-05-03T19:32:33+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```
    """
    validators = clearskies.configs.validators.Validators(default=[])

    """
    Actions to take during the pre-save step of the save process if the column has changed during the active save operation.

    Pre-save happens before the data is persisted to the backend.  Actions/callables in
    this step must return a dictionary.  The data in the dictionary will be included in the save operation.
    Since the save hasn't completed, any data in the model itself reflects the model before the save
    operation started.  Actions in the pre-save step must **NOT** make any changes directly, but should **ONLY**
    return modified data for the save operation.  In addition, they must be idempotent - they should always return
    the same value when called with the same data.  This is because clearskies can call them more than once.  If
    a pre-save hook changes the save data, then clearskies will call all the pre-save hooks again in case this
    new data needs to trigger further changes.  Stateful changes should be reserved for the post_save or save_finished stages.

    Callables and actions can request any dependencies provided by the DI system.  In addition, they can request
    two named parameters:

     1. `model` - the model involved in the save operation
     2. `data` - the new data being saved

    The key here is that the defined actions will be invoked regardless of how the save happens.  Whether the
    model.save() function is called directly or the model is creatd/modified via an endpoint, your business logic
    will always be executed.  This makes for easy reusability and consistency throughout your application.

    Here's an example where we want to record a timestamp anytime an order status becomes a particular value:

    ```python
    import clearskies

    class Order(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        status = clearskies.columns.Select(
            ["Open", "On Hold", "Fulfilled"],
            on_change_pre_save=[
                lambda data, utcnow: {"fulfilled_at": utcnow} if data["status"] == "Fulfilled" else {},
            ],
        )
        fulfilled_at = clearskies.columns.Datetime()

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            model_class=Order,
            writeable_column_names=["status"],
            readable_column_names=["id", "status", "fulfilled_at"],
        ),
    )
    wsgi()
    ```

    You can then see the difference depending on what you set the status to:

    ```bash
    $ curl http://localhost:8080 -d '{"status":"Open"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "a732545f-51b3-4fd0-a6cf-576cf1b2872f",
            "status": "Open",
            "fulfilled_at": null
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl http://localhost:8080 -d '{"status":"Fulfilled"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "c288bf43-2246-48e4-b168-f40cbf5376df",
            "status": "Fulfilled",
            "fulfilled_at": "2025-05-04T02:32:56+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    ```

    """
    on_change_pre_save = clearskies.configs.actions.Actions(default=[])

    """
    Actions to take during the post-save step of the process if the column has changed during the active save.

    Post-save happens after the data is persisted to the backend but before the full save process has finished.
    Since the data has been persisted to the backend, any data returned by the callables/actions will be ignored.
    If you need to make data changes you'll have to execute a separate save operation.
    Since the save hasn't finished, the model is not yet updated with the new data, and
    any data you fetch out of the model will refelect the data in the model before the save started.

    Callables and actions can request any dependencies provided by the DI system.  In addition, they can request
    three named parameters:

     1. `model` - the model involved in the save operation
     2. `data` - the new data being saved
     3. `id` - the id of the record being saved

    Here's an example of using a post-save action to record a simple audit trail when the order status changes:

    ```python
    import clearskies

    class Order(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        status = clearskies.columns.Select(
            ["Open", "On Hold", "Fulfilled"],
            on_change_post_save=[
                lambda model, data, order_histories: order_histories.create({
                    "order_id": model.latest("id", data),
                    "event": "Order status changed to " + data["status"]
                }),
            ],
        )

    class OrderHistory(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        event = clearskies.columns.String()
        order_id = clearskies.columns.BelongsToId(Order)

        # include microseconds in the created_at time so that we can sort our example by created_at
        # and they come out in order (since, for our test program, they will all be created in the same second).
        created_at = clearskies.columns.Created(date_format="%Y-%m-%d %H:%M:%S.%f")

    def test_post_save(orders: Order, order_histories: OrderHistory):
        my_order = orders.create({"status": "Open"})
        my_order.status = "On Hold"
        my_order.save()
        my_order.save({"status": "Open"})
        my_order.save({"status": "Fulfilled"})
        return order_histories.where(OrderHistory.order_id.equals(my_order.id)).sort_by("created_at", "asc")

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            test_post_save,
            model_class=OrderHistory,
            return_records=True,
            readable_column_names=["id", "event", "created_at"],
        ),
        classes=[Order, OrderHistory],
    )
    cli()
    ```

    Note that in our `on_change_post_save` lambda function, we use `model.latest("id", data)`.  We can't just use
    `data["id"]` because `data` is a dictionary containing the information present in the save.  During the create
    operation `data["id"]` will be populated, but during the subsequent edit operations it won't be - only the status
    column is changing.  `model.latest("id", data)` is basically just short hand for: `data.get("id", model.id)`.
    On the other hand, we can just use `data["status"]` because the `on_change` hook is attached to the status field,
    so it will only fire when status is being changed, which means that the `status` key is guaranteed to be in
    the dictionary when the lambda is executed.

    Finally, the post-save action has a named parameter called `id`, so in this specific case we could use:

    ```python
    lambda data, id, order_histories: order_histories.create("order_id": id, "event": data["status"])
    ```

    When we execute the above script it will return something like:

    ```json
    {
        "status": "success",
        "error": "",
        "data": [
            {
            "id": "c550d714-839b-4f25-a9e1-bd7e977185ff",
            "event": "Order status changed to Open",
            "created_at": "2025-05-04T14:09:42.960119+00:00"
            },
            {
            "id": "f393d7b0-da21-4117-a7a4-0359fab802bb",
            "event": "Order status changed to On Hold",
            "created_at": "2025-05-04T14:09:42.960275+00:00"
            },
            {
            "id": "5b528a10-4a08-47ae-938c-fc7067603f8e",
            "event": "Order status changed to Open",
            "created_at": "2025-05-04T14:09:42.960395+00:00"
            },
            {
            "id": "91f77a88-1c38-49f7-aa1e-7f97bd9f962f",
            "event": "Order status changed to Fulfilled",
            "created_at": "2025-05-04T14:09:42.960514+00:00"
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```

    """
    on_change_post_save = clearskies.configs.actions.Actions(default=[])

    """
    Actions to take during the save-finished step of the save process if the column has changed in the save.

    Save-finished happens after the save process has completely finished and the model is updated with
    the final data.  Any data returned by these actions will be ignored, since the save has already finished.
    If you need to make data changes you'll have to execute a separate save operation.

    Callables and actions can request any dependencies provided by the DI system.  In addition, they can request
    the following parameter:

     1. `model` - the model involved in the save operation

    Unlike pre_save and post_save, `data` is not provided because this data has already been merged into the
    model.  If you need some context from the completed save operation, use methods like `was_changed` and `previous_value`.
    """
    on_change_save_finished = clearskies.configs.actions.Actions(default=[])

    """
    Use in conjunction with `created_by_source_type` to have this column automatically populated by data from an HTTP request.

    So, for instance, setting `created_by_source_type` to `authorization_data` and setting this to `email`
    will result in the email value from the authorization data being persisted into this column when the
    record is saved.

    NOTE: this is sometimes best set as a column override on an endpoint, rather than directly
    on the model itself.  The reason is because the authorization data and header information is typically
    only available during an HTTP request, so if you set this on the model level, you'll get an error
    if you try to make saves to the model in a context where authorization data and/or headers don't exist.

    See created_by_source_type for usage examples.
    """
    created_by_source_key = clearskies.configs.string.String(default="")

    """
    Use in conjunction with `created_by_source_key` to have this column automatically populated by data from ann HTTP request.

    So, for instance, setting this to `authorization_data` and setting `created_by_source_key` to `email`
    will result in the email value from the authorization data being persisted into this column when the
    record is saved.

    NOTE: this is sometimes best set as a column override on an endpoint, rather than directly
    on the model itself.  The reason is because the authorization data and header information is typically
    only available during an HTTP request, so if you set this on the model level, you'll get an error
    if you try to make saves to the model in a context where authorization data and/or headers don't exist.

    Here's an example:

    ```python
    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        account_id = clearskies.columns.String(
            created_by_source_type="routing_data",
            created_by_source_key="account_id",
        )

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            User,
            readable_column_names=["id", "account_id", "name"],
            writeable_column_names=["name"],
            url="/:account_id",
        ),
    )
    wsgi()
    ```

    Note that `created_by_source_type` is `routing_data` and `created_by_source_key` is `account_id`.
    This means that the endpoint that creates this record must have a routing parameter named `account_id`.
    Naturally, our endpoint has a url of `/:account_id`, and so the parameter provided by the uesr gets
    reflected into the save.

    ```bash
    $ curl http://localhost:8080/1-2-3-4 -d '{"name":"Bob"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "250ed725-d940-4823-aa9d-890be800404a",
            "account_id": "1-2-3-4",
            "name": "Bob"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    """
    created_by_source_type = clearskies.configs.select.Select(
        ["authorization_data", "http_header", "routing_data", ""], default=""
    )

    """
    If True, and the key requested via created_by_source_key doesn't exist in the designated source, an error will be raised.
    """
    created_by_source_strict = clearskies.configs.boolean.Boolean(default=True)

    """ The model class this column is associated with. """
    model_class = clearskies.configs.Schema()

    """ The name of this column. """
    name = clearskies.configs.string.String()

    """
    Simple flag to denote if the column is unique or not.

    This is an internal cache.  Use column.is_unique instead.
    """
    _is_unique = False

    """
    Specify if this column has additional functionality to solve the n+1 problem.

    Relationship columns may fetch data from additional tables when outputting results, but by default they
    end up making an additional query for every record (in order to grab related data).  This is called the
    n+1 problem - a query may fetch 10 records, and then make 10 individual additional queries to select
    related data for each record (which obviously hampers performance).  The solution to this (when using
    sql-like backends) is to add additional joins to the original query so that the data can all be fetched
    at once.  Columns that are subject to this issue can set this flag to True and then  define the
    `configure_n_plus_one` method to add the necessary joins.  This method will be called as needed.
    """
    wants_n_plus_one = False

    """
    Simple flag to denote if the column is required or not.

    This is an internal cache.  Use column.is_required instead.
    """
    _is_required = False

    """
    The list of allowed search operators for this column.

    All the various search methods reference this list.  The idea is that a column can just fill out this list
    instead of having to override all the methods.
    """
    _allowed_search_operators = ["<=>", "!=", "<=", ">=", ">", "<", "=", "in", "is not null", "is null", "like"]

    """
    The class to use when documenting this column
    """
    auto_doc_class: type[AutoDocSchema] = AutoDocString

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        default: str | None = None,
        setable: str | Callable[..., str] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def get_model_columns(self):
        """Return the columns or the model this column is attached to."""
        return self.model_class.get_columns()

    def finalize_configuration(self, model_class: type[Schema], name: str) -> None:
        """
        Finalize and check the configuration.

        This is an external trigger called by the model class when the model class is ready.
        The reason it exists here instead of in the constructor is because some columns are tightly
        connected to the model class, and can't validate configuration until they know what the model is.
        Therefore, we need the model involved, and the only way for a property to know what class it is
        in is if the parent class checks in (which is what happens here).
        """
        self.model_class = model_class
        self.name = name
        self.finalize_and_validate_configuration()

    def from_backend(self, value):
        """
        Take the backend representation and returns a python representation.

        For instance, for an SQL date field, this will return a Python DateTime object
        """
        return str(value)

    def to_backend(self, data: dict[str, Any]) -> dict[str, Any]:
        """
        Make any changes needed to save the data to the backend.

        This typically means formatting changes - converting DateTime objects to database
        date strings, etc...
        """
        if self.name not in data:
            return data

        return {**data, self.name: str(data[self.name])}

    @overload
    def __get__(self, instance: None, cls: type) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type):
        pass

    def __get__(self, instance, cls):
        if instance is None:
            # Normally this gets filled in when the model is initialized.  However, the condition builders (self.equals, etc...)
            # can be called from the class directly, before the model is initialized and everything is populated.  This
            # can cause trouble, but by filling in the model class we can give enough information for them to get the
            # job done.  They have a special flow for this, we just have to provide the model class (and the __get__
            # function is always called, so this fixes it).
            self.model_class = cls
            return self

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            instance.get_columns()

        if self.name not in instance._data:
            return None  # type: ignore

        if self.name not in instance._transformed_data:
            instance._transformed_data[self.name] = self.from_backend(instance._data[self.name])

        return instance._transformed_data[self.name]

    def __set__(self, instance: Model, value) -> None:
        instance._next_data[self.name] = value

    def finalize_and_validate_configuration(self):
        super().finalize_and_validate_configuration()

        if self.setable is not None and self.created_by_source_type:
            raise ValueError(
                "You attempted to set both 'setable' and 'created_by_source_type', but these configurations are mutually exclusive.  You can only set one for a given column"
            )

        if (self.created_by_source_type and not self.created_by_source_key) or (
            not self.created_by_source_type and self.created_by_source_key
        ):
            raise ValueError(
                "You only set one of 'created_by_source_type' and 'created_by_source_key'.  You have to either set both of them (which enables the 'created_by' feature of the column) or you must set neither of them."
            )

    @property
    def is_unique(self) -> bool:
        """Return True/False to denote if this column should always have unique values."""
        if self._is_unique is None:
            self._is_unique = any([validator.is_unique for validator in self.validators])
        return self._is_unique

    @property
    def is_required(self):
        """Return True/False to denote if this column should is required."""
        if self._is_required is None:
            self._is_required = any([validator.is_required for validator in self.validators])
        return self._is_required

    def additional_write_columns(self, is_create=False) -> dict[str, Self]:
        """
        Return any additional columns that should be included in write operations.

        Some column types, and some validation requirements, necessitate the presence of additional
        columns in the save operation.  This function adds those in so they can be included in the
        API call.
        """
        additional_write_columns: dict[str, Self] = {}
        for validator in self.validators:
            if not isinstance(validator, Validator):
                continue
            additional_write_columns = {
                **additional_write_columns,
                **validator.additional_write_columns(is_create=is_create),  # type: ignore
            }
        return additional_write_columns

    def to_json(self, model: clearskies.model.Model) -> dict[str, Any]:
        """Grabs the column out of the model and converts it into a representation that can be turned into JSON."""
        return {self.name: self.__get__(model, model.__class__)}

    def input_errors(self, model: clearskies.model.Model, data: dict[str, Any]) -> dict[str, Any]:
        """
        Check the given dictionary of data for any possible input errors.

        This accepts all the data being saved, and not just the value for this column.  The reason is because
        some input valdiation flows require more than one piece of data.  For instance, a user may be asked
        to type a specific piece of input more than once to minimize the chance of typos, or a user may
        have to provide their password when changing security-related columns.

        This also returns a dictionary, rather than an error message, so that a column can also return an error
        message for more than one column at a time if needed.

        If there are no input errors then this should simply return an empty dictionary.

        This method calls `self.input_error_for_value` and then also calls all the validators attached to the
        column so, if you're building your own column and have some specific input validation you need to do,
        you probably want to extend `input_error_for_value` as that is the one intended checks for a column type.

        Note: this is not called when you directly invoke the `save`/`create` method of a model.  This is only
        used by handlers when processing user input (e.g. API calls).
        """
        if self.name in data and data[self.name]:
            error = self.input_error_for_value(data[self.name])
            if error:
                return {self.name: error}

        for validator in self.validators:
            if hasattr(validator, "injectable_properties"):
                validator.injectable_properties(self.di)

            error = validator(model, self.name, data)
            if error:
                return {self.name: error}

        return {}

    def check_search_value(
        self, value: str, operator: str | None = None, relationship_reference: str | None = None
    ) -> str:
        """
        Check if the given value is an allowed value.

        This is called by the search operation in the various API-related handlers to validate a search value.

        Generally, this just defers to self.input_error_for_value, but it is a separate method in case you
        need to change your input validation logic specifically when checking a search value.
        """
        return self.input_error_for_value(value, operator=operator)

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        """
        Check if the given value is an allowed value.

        This method is intended for checks that are specific to the column type (e.g. this is where an
        email column checks that the value is an actual email, or a datetime column checks for a valid
        datetime).  The `input_errors` method does a bit more, so in general it's easier to extend this one.

        This method is passed in the value to check.  It should return a string.  If the data is valid,
        then return an empty string.  Otherwise return a human-readable error message.

        At times an operator will be passed in.  This is used when the user is searching instead of saving.
        In this case, the check can vary depending on the operator.  For instance, if it's a wildcard search
        then an email field only has to verify the type is a string (since the user may have only entered
        the beginning of an email address), but if it's an exact search then you would expect the value to be
        an actual email.

        Note: this is not called when you directly invoke the `save`/`create` method of a model.  This is only
        used by handlers when processing user input (e.g. API calls).
        """
        return ""

    def pre_save(self, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        """
        Make any necessary changes to the data before starting the save process.

        The difference between this and to_backend is that to_backend only affects
        the data as it is going into the database, while this affects the data that will get persisted
        in the object as well.  So for instance, for a "created" field, pre_save may fill in the current
        date with a Python DateTime object when the record is being saved, and then to_backend may
        turn that into an SQL-compatible date string.

        Note: this is called during the `pre_save` step in the lifecycle of the save process.  See the
        model class for more details.
        """
        if not model and self.created_by_source_type:
            data[self.name] = self._extract_value_from_source_type()
        if self.setable:
            if callable(self.setable):
                input_output = self.di.build("input_output", cache=True)
                data[self.name] = self.di.call_function(
                    self.setable, data=data, model=model, **input_output.get_context_for_callables()
                )
            else:
                data[self.name] = self.setable
        if not model and self.default and self.name not in data:
            data[self.name] = self.default
        if self.on_change_pre_save and model.is_changing(self.name, data):
            data = self.execute_actions_with_data(self.on_change_pre_save, model, data)
        return data

    def post_save(self, data: dict[str, Any], model: clearskies.model.Model, id: int | str) -> None:
        """
        Make any changes needed after persisting data to the backend.

        This lives in the `post_save` hook of the save lifecycle.  See the model class for more details.
        `data` is the data dictionary being saved.  `model` is obviously the model object that initiated the
        save.

        This happens after the backend is updated but before the model is updated.  Therefore, You can tell the
        difference between a create operation and an update operation by checking if the model exists: `if model`.
        For a create operation, the model will be empty (it evaluates to False).  The opposite is true for an update
        operation.

        Any return value will be ignored.  If you need to make additional changes in the backend, you
        have to execute a new save operation.
        """
        if self.on_change_post_save and model.is_changing(self.name, data):
            self.execute_actions_with_data(
                self.on_change_post_save,
                model,
                data,
                id=id,
                context="on_change_post_save",
                require_dict_return_value=False,
            )

    def save_finished(self, model: clearskies.model.Model) -> None:
        """
        Make any necessary changes needed after a save has completely finished.

        This is typically used for actions set by the developer.   Column-specific behavior usually lives in
        `pre_save` or `post_save`.  See the model class for more details about the various lifecycle hooks during
        a save.
        """
        if self.on_change_save_finished and model.was_changed(self.name):
            self.execute_actions(self.on_change_save_finished, model)

    def pre_delete(self, model):
        """Make any changes needed to the data before starting the delete process."""
        pass

    def post_delete(self, model):
        """Make any changes needed to the data before finishing the delete process."""
        pass

    def _extract_value_from_source_type(self) -> Any:
        """For columns with `created_by_source_type` set, this fetches the appropriate value from the request."""
        input_output = self.di.build("input_output", cache=True)
        source_type = self.created_by_source_type
        if source_type == "authorization_data":
            data = input_output.authorization_data
        elif source_type == "http_header":
            data = input_output.request_headers
        elif source_type == "routing_data":
            data = input_output.routing_data

        if self.created_by_source_key not in data and self.created_by_source_strict:
            raise ValueError(
                f"Column '{self.name}' is configured to load the key named '{self.created_by_source_key}' from "
                + f"the {self.created_by_source_type}', but this key was not present in the request."
            )

        return data.get(self.created_by_source_key, "N/A")

    def execute_actions_with_data(
        self,
        actions: list[clearskies.typing.action],
        model: clearskies.model.Model,
        data: dict[str, Any],
        id: int | str | None = None,
        context: str = "on_change_pre_save",
        require_dict_return_value: bool = True,
    ) -> dict[str, Any]:
        """Execute a given set of actions and expects data to be both provided and returned."""
        input_output = self.di.build("input_output", cache=True)
        for index, action in enumerate(actions):
            new_data = self.di.call_function(
                action,
                **{
                    **input_output.get_context_for_callables(),
                    **{
                        "model": model,
                        "data": data,
                        "id": id,
                    },
                },
            )
            if not isinstance(new_data, dict):
                if require_dict_return_value:
                    raise ValueError(
                        f"Return error for action #{index + 1} in 'on_change_pre_save' for column '{self.name}' in model '{self.model_class.__name__}': this action must return a dictionary but returned an object of type '{new_data.__class__.__name__}' instead"
                    )
                else:
                    return new_data
            data = {
                **data,
                **new_data,
            }
        return data

    def execute_actions(
        self,
        actions: list[clearskies.typing.action],
        model: clearskies.model.Model,
    ) -> None:
        """Execute a given set of actions."""
        input_output = self.di.build("input_output", cache=True)
        for action in actions:
            self.di.call_function(action, model=model, **input_output.get_context_for_callables())

    def values_match(self, value_1, value_2):
        """
        Compare two values to see if they are the same.

        This is mainly used to compare incoming data with old data to determine if a column has changed.

        Note that these checks shouldn't make any assumptions about whether or not data has gone through the
        to_backend/from_backend functions.  For instance, a datetime field may find one value has a date
        that is formatted as a string, and the other as a DateTime object.  Plan appropriately.
        """
        return value_1 == value_2

    def add_search(
        self, model: clearskies.model.Model, value: str, operator: str = "", relationship_reference: str = ""
    ) -> clearskies.model.Model:
        return model.where(self.condition(operator, value))

    def build_condition(self, value: str, operator: str = "", column_prefix: str = ""):
        """
        Build condition for the read (and related) handlers to turn user input into a condition.

        Note that this may look like it is vulnerable to SQLi, but it isn't.  These conditions aren't passed directly
        into a query.  Rather, they are parsed by the condition parser before being sent into the backend.
        The condition parser can safely reconstruct the original pieces, and the backend can then use the data
        safely (and remember, the backend may not be an SQL anyway)

        As a result, this is perfectly safe for any user input, assuming normal system flow.

        That being said, this should probably be replaced by self.condition()...
        """
        if not operator:
            operator = "="
        if operator.lower() == "like":
            return f"{column_prefix}{self.name} LIKE '%{value}%'"
        return f"{column_prefix}{self.name}{operator}{value}"

    def is_allowed_operator(
        self,
        operator: str,
        relationship_reference: str = "",
    ):
        """Process user data to decide if the end-user is specifying an allowed operator."""
        return operator.lower() in self._allowed_search_operators

    def n_plus_one_add_joins(
        self, model: clearskies.model.Model, column_names: list[str] = []
    ) -> clearskies.model.Model:
        """Add any additional joins to solve the N+1 problem."""
        return model

    def n_plus_one_join_table_alias_prefix(self):
        """
        Create a table alias to use with joins for n+1 solutions.

        When joining tables in for n+1 solutions, you can't just do a SELECT * on the new table, because that
        often results in duplicate column names.  A solution that generally works across the board is to select
        specific columns from the joined table and alias them, adding a common prefix.  Then, the data from the
        joined table can be reconstructed automatically by finding all columns with that prefix (and then removing
        the prefix).  This function returns that prefix for that alias.

        Now, technically this isn't function isn't used at all by the base class, so this definition is fairly
        pointless.  It isn't marked as an abstract method because most model columns don't need it either.
        Rather, this function is here mostly for documentation so it's easier to understand how to implement
        support for n+1 solutions when needed.  See the belongs_to column for a full implementation reference.
        """
        return "join_table_" + self.name

    def add_join(self, model: Model) -> Model:
        return model

    def where_for_request(
        self,
        model: clearskies.model.Model,
        routing_data: dict[str, str],
        authorization_data: dict[str, Any],
        input_output,
    ) -> clearskies.model.Model:
        """
        Create a hook to automatically apply filtering whenever the column makes an appearance in a get/update/list/search handler.

        This hook is called by all the handlers that execute queries, so if your column needs to automatically
        do some filtering whenever the model shows up in an API endpoint, this is the place for it.
        """
        return model

    def name_for_building_condition(self) -> str:
        if self._config and "name" in self._config:
            return self.name

        if not self._config or not self._config.get("model_class"):
            raise ValueError(
                f"A condition builder was called but the model class isn't set.  This means that the __get__ method for column class {self.__class__.__name__} forgot to set `self.model_class = cls`"
            )

        for attribute_name in dir(self.model_class):
            if id(getattr(self.model_class, attribute_name)) != id(self):
                continue
            self.name = attribute_name
            break

        return self.name

    def equals(self, value) -> Condition:
        name = self.name_for_building_condition()
        if "=" not in self._allowed_search_operators:
            raise ValueError(f"An 'equals search' is not allowed for '{self.model_class.__name__}.{name}'.")
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, "=", [value])

    def spaceship(self, value) -> Condition:
        name = self.name_for_building_condition()
        if "<=>" not in self._allowed_search_operators:
            raise ValueError(f"A 'spaceship' search is not allowed for '{self.model_class.__name__}.{name}'.")
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, "<=>", [value])

    def not_equals(self, value) -> Condition:
        name = self.name_for_building_condition()
        if "!=" not in self._allowed_search_operators:
            raise ValueError(f"A 'not equals' search is not allowed for '{self.model_class.__name__}.{name}'.")
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, "!=", [value])

    def less_than_equals(self, value) -> Condition:
        name = self.name_for_building_condition()
        if "<=" not in self._allowed_search_operators:
            raise ValueError(f"A 'less than or equals' search is not allowed for '{self.model_class.__name__}.{name}'.")
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, "<=", [value])

    def greater_than_equals(self, value) -> Condition:
        name = self.name_for_building_condition()
        if ">=" not in self._allowed_search_operators:
            raise ValueError(
                f"A 'greater than' or equals search is not allowed for '{self.model_class.__name__}.{name}'."
            )
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, ">=", [value])

    def less_than(self, value) -> Condition:
        name = self.name_for_building_condition()
        if "<" not in self._allowed_search_operators:
            raise ValueError(f"A 'less than' search is not allowed for '{self.model_class.__name__}.{name}'.")
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, "<", [value])

    def greater_than(self, value) -> Condition:
        name = self.name_for_building_condition()
        if ">" not in self._allowed_search_operators:
            raise ValueError(f"A 'greater than' search is not allowed for '{self.model_class.__name__}.{name}'.")
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, ">", [value])

    def is_in(self, values) -> Condition:
        name = self.name_for_building_condition()
        if "in" not in self._allowed_search_operators:
            raise ValueError(f"An 'in' search is not allowed for '{self.model_class.__name__}.{name}'.")
        if not isinstance(values, list):
            raise TypeError("You must pass a list in to column.is_in")
        final_values = []
        for value in values:
            final_values.append(self.to_backend({name: value}).get(name))
        return ParsedCondition(name, "in", final_values)  # type: ignore

    def is_not_null(self) -> Condition:
        name = self.name_for_building_condition()
        if "is not null" not in self._allowed_search_operators:
            raise ValueError(f"An 'is not null' search is not allowed for '{self.model_class.__name__}.{name}'.")
        return ParsedCondition(name, "is not null", [])

    def is_null(self) -> Condition:
        name = self.name_for_building_condition()
        if "is null" not in self._allowed_search_operators:
            raise ValueError(f"An 'is null' search is not allowed for '{self.model_class.__name__}.{name}'.")
        return ParsedCondition(name, "is null", [])

    def like(self, value) -> Condition:
        name = self.name_for_building_condition()
        if "like" not in self._allowed_search_operators:
            raise ValueError(f"A 'like' search is not allowed for '{self.model_class.__name__}.{name}'.")
        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, "like", [value])

    def condition(self, operator: str, value) -> Condition:
        name = self.name_for_building_condition()
        operator = operator.lower()
        if operator not in self._allowed_search_operators:
            raise ValueError(f"The operator '{operator}' is not allowed for '{self.model_class.__name__}.{name}'.")

        # the search methods are more or less identical except:
        if operator == "in":
            return self.is_in(value)

        value = self.to_backend({name: value}).get(name)
        return ParsedCondition(name, operator, [value])

    def is_allowed_search_operator(self, operator: str, relationship_reference: str = "") -> bool:
        return operator in self._allowed_search_operators

    def allowed_search_operators(self, relationship_reference: str = ""):
        return self._allowed_search_operators

    def join_table_alias(self) -> str:
        raise NotImplementedError("Ooops, I don't support joins")

    def documentation(self, name=None, example=None, value=None) -> list[AutoDocSchema]:
        return [self.auto_doc_class(name if name is not None else self.name, example=example, value=value)]



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import clearskies.configs
from clearskies.validator import Validator

if TYPE_CHECKING:
    import clearskies.model


class MaximumLength(Validator):
    maximum_length = clearskies.configs.Integer(required=True)

    def __init__(self, maximum_length: int):
        self.maximum_length = maximum_length
        self.finalize_and_validate_configuration()

    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        # we won't check anything for missing values (columns should be required if that is an issue)
        if not data.get(column_name):
            return ""
        if len(data[column_name]) <= self.maximum_length:
            return ""
        return f"'{column_name}' must be at most {self.maximum_length} characters long."



from __future__ import annotations

from typing import Any

import clearskies.typing
from clearskies import configs, parameters_to_properties
from clearskies.column import Column
from clearskies.columns.has_many import HasMany


class Audit(HasMany):
    """
    Enables auditing for a model.

    Specify the audit class to use and attach this column to your model. Everytime the model is created/updated/deleted,
    the audit class will record the action and the changes.  Your audit model must have the following columns:

    | Name        | type     |
    |-------------|----------|
    | class       | str      |
    | resource_id | str      |
    | action      | str      |
    | data        | json     |
    | created_at  | created  |

    The names are not currently adjustable.

     1. Class is a string that records the name of the class that the action happened for.  This allows you to use
        the same audit class for multiple, different, resources.
     2. resource_id is the id of the record which the audit entry is for.
     3. Action is the actual action taken (create/update/delete)
     4. Data is a serialized record of what columns in the record were changed (both their previous and new values)
     5. The time the audit record was created
    """

    """ The model class for the destination that will store the audit data. """
    audit_model_class = configs.ModelClass(required=True)

    """
    A list of columns that shouldn't be copied into the audit record.

    To be clear, these are columns from the model class that the audit column is attached to.
    If only excluded columns are updated then no audit record will be created.
    """
    exclude_columns = configs.ModelColumns()

    """
    A list of columns that should be masked when copied into the audit record.

    With masked columns a generic value is placed in the audit record (e.g. XXXXX) which denotes that
    the column was changed, but it does not record either old or new values.
    """
    mask_columns = configs.ModelColumns()

    """ Columns from the child table that should be included when converting this column to JSON. """
    readable_child_columns = configs.ReadableModelColumns(
        "audit_model_class", default=["resource_id", "action", "data", "created_at"]
    )

    """
    Since this column is always populated automatically, it is never directly writeable.
    """
    is_writeable = configs.Boolean(default=False)
    is_searchable = configs.Boolean(default=False)
    _descriptor_config_map = None
    _parent_columns: dict[str, Column] | None

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        audit_model_class,
        exclude_columns: list[str] = [],
        mask_columns: list[str] = [],
        foreign_column_name: str | None = None,
        readable_child_columns: list[str] = [],
        where: clearskies.typing.condition | list[clearskies.typing.condition] = [],
        default: str | None = None,
        is_readable: bool = True,
        is_temporary: bool = False,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        self.child_model_class = self.audit_model_class

    def save_finished(self, model):
        super().save_finished(model)
        old_data: dict[str, Any] = model._previous_data
        new_data: dict[str, Any] = model.get_raw_data()
        exclude_columns = self.exclude_columns
        mask_columns = self.mask_columns
        model_columns = self.get_model_columns()

        if not old_data:
            create_data: dict[str, Any] = {}
            for key in new_data.keys():
                if key in exclude_columns:
                    continue
                if key in model_columns:
                    column_data = model_columns[key].to_json(model)
                else:
                    column_data = {key: new_data[key]}

                create_data = {
                    **create_data,
                    **column_data,
                }
                if key in mask_columns and key in create_data:
                    create_data[key] = "****"
            self.record(model, "create", data=create_data)
            return

        # note that this is fairly simple logic to get started.  It's not going to detect changes that happen
        # in other "tables".  For instance, disconnecting a record by deleting an entry in a many-to-many relationship
        # won't be picked up by this.
        old_model = model.empty_model()
        old_model.data = old_data
        from_data: dict[str, Any] = {}
        to_data: dict[str, Any] = {}
        for column, new_value in new_data.items():
            if column in exclude_columns or column not in old_data:
                continue
            if old_data[column] == new_value:
                continue
            from_data = {
                **from_data,
                **(
                    model_columns[column].to_json(old_model)
                    if column in model_columns
                    else {column: old_data.get(column)}
                ),
            }
            to_data = {
                **to_data,
                **(
                    model_columns[column].to_json(model)
                    if column in model_columns
                    else {column: model._data.get(column)}
                ),
            }
            if column in mask_columns and column in to_data:
                to_data[column] = "****"
                from_data[column] = "****"
        if not from_data and not to_data:
            return

        self.record(
            model,
            "update",
            data={
                "from": from_data,
                "to": to_data,
            },
        )

    def post_delete(self, model):
        super().post_delete(model)
        exclude_columns = self.exclude_columns
        model_columns = self.get_model_columns()
        mask_columns = self.mask_columns

        final_data: dict[str, Any] = {}
        for key in model._data.keys():
            if key in exclude_columns:
                continue
            final_data = {
                **final_data,
                **(model_columns[key].to_json(model) if key in model_columns else {key: model.data.get(key)}),
            }

        for key in mask_columns:
            if key not in final_data:
                continue
            final_data[key] = "****"

        self.child_model.create(
            {
                "class": self.model_class.__name__,
                "resource_id": model.get(self.model_class.id_column_name),
                "action": "delete",
                "data": final_data,
            }
        )

    @property
    def parent_columns(self) -> dict[str, Column]:
        if self._parent_columns == None:
            self._parent_columns = self.di.build(self.model_class, cache=True).columns()
        return self._parent_columns

    def record(self, model, action, data=None, record_data=None):
        audit_data = {
            "class": self.model_class.__name__,
            "resource_id": model.get(self.model_class.id_column_name),
            "action": action,
        }
        if data is not None:
            audit_data["data"] = data
        if record_data is not None:
            audit_data = {
                **audit_data,
                **record_data,
            }

        self.child_model.create(audit_data)



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable, Self, overload

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.autodoc.schema import Array as AutoDocArray
from clearskies.autodoc.schema import String as AutoDocString
from clearskies.column import Column
from clearskies.functional import string

if TYPE_CHECKING:
    from clearskies import Column, Model


class ManyToManyIds(Column):
    """
    A column that represents a many-to-many relationship.

    This is different from belongs to/has many because with those, every child has only one parent.  With a many-to-many
    relationship, both models can have multiple relatives from the other model class.  In order to support this, it's necessary
    to have a third model (the pivot model) that records the relationships.  In general this table just needs three
    columns: it's own id, and then one column for each other model to store the id of the related records.
    You can specify the names of these columns but it also follows the standard naming convention by default:
    take the class name, convert it to snake case, and append `_id`.

    Note, there is a variation on this (`ManyToManyIdsWithData`) where additional data is stored in the pivot table
    to record information about the relationship.

    This column is writeable.  You would set it to a list of ids from the related model that denotes which
    records it is related to.

    The following example shows usage.  Normally the many-to-many column exists for both related models, but in this
    specific example it only exists for one of the models.  This is done so that the example can fit in a single file
    and therefore be easy to demonstrate.  In order to have both models reference eachother, you have to use model
    references to avoid circular imports.  There are examples of doing this in the `BelongsTo` column class.

    ```python
    import clearskies


    class ThingyToWidget(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        # these could also be belongs to relationships, but the pivot model
        # is rarely used directly, so I'm being lazy to avoid having to use
        # model references.
        thingy_id = clearskies.columns.String()
        widget_id = clearskies.columns.String()


    class Thingy(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()


    class Widget(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        thingy_ids = clearskies.columns.ManyToManyIds(
            related_model_class=Thingy,
            pivot_model_class=ThingyToWidget,
        )
        thingies = clearskies.columns.ManyToManyModels("thingy_ids")


    def my_application(widgets: Widget, thingies: Thingy):
        thing_1 = thingies.create({"name": "Thing 1"})
        thing_2 = thingies.create({"name": "Thing 2"})
        thing_3 = thingies.create({"name": "Thing 3"})
        widget = widgets.create({
            "name": "Widget 1",
            "thingy_ids": [thing_1.id, thing_2.id],
        })

        # remove an item by saving without it's id in place
        widget.save({"thingy_ids": [thing.id for thing in widget.thingies if thing.id != thing_1.id]})

        # add an item by saving and adding the new id
        widget.save({"thingy_ids": [*widget.thingy_ids, thing_3.id]})

        return widget.thingies


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            my_application,
            model_class=Thingy,
            return_records=True,
            readable_column_names=["id", "name"],
        ),
        classes=[Widget, Thingy, ThingyToWidget],
    )

    if __name__ == "__main__":
        cli()
    ```

    And when executed:

    ```json
    {
        "status": "success",
        "error": "",
        "data": [
            {"id": "741bc838-c694-4624-9fc2-e9032f6cb962", "name": "Thing 2"},
            {"id": "1808a8ef-e288-44e6-9fed-46e3b0df057f", "name": "Thing 3"},
        ],
        "pagination": {},
        "input_errors": {},
    }
    ```

    Of course, you can also create or remove individual relationships by using the pivot model directly,
    as shown in these partial code snippets:

    ```python
    def add_items(thingy_to_widgets):
        thingy_to_widgets.create({
            "thingy_id": "some_id",
            "widget_id": "other_id",
        })


    def remove_item(thingy_to_widgets):
        thingy_to_widgets.where("thingy_id=some_id").where("widget_id=other_id").first().delete()
    ```
    """

    """ The model class for the model that we are related to. """
    related_model_class = configs.ModelClass(required=True)

    """ The model class for the pivot table - the table used to record connections between ourselves and our related table. """
    pivot_model_class = configs.ModelClass(required=True)

    """
    The name of the column in the pivot table that contains the id of records from the model with this column.

    A default name is created by taking the model class name, converting it to snake case, and then appending `_id`.
    If you name your columns according to this standard then you don't have to specify this column name.
    """
    own_column_name_in_pivot = configs.ModelToIdColumn(model_column_config_name="pivot_model_class")

    """
    The name of the column in the pivot table that contains the id of records from the related table.

    A default name is created by taking the name of the related model class, converting it to snake case, and then
    appending `_id`. If you name your columns according to this standard then you don't have to specify this column
    name.
    """
    related_column_name_in_pivot = configs.ModelToIdColumn(
        model_column_config_name="pivot_model_class", source_model_class_config_name="related_model_class"
    )

    """ The name of the pivot table."""
    pivot_table_name = configs.ModelDestinationName("pivot_model_class")

    """ The list of columns to be loaded from the related models when we are converted to JSON. """
    readable_related_column_names = configs.ReadableModelColumns("related_model_class")

    default = configs.StringList(default=None)  #  type: ignore
    setable = configs.StringListOrCallable(default=None)  #  type: ignore
    is_searchable = configs.Boolean(default=False)
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        related_model_class,
        pivot_model_class,
        own_column_name_in_pivot: str = "",
        related_column_name_in_pivot: str = "",
        readable_related_column_names: list[str] = [],
        default: list[str] = [],
        setable: list[str] | Callable[..., list[str]] = [],
        is_readable: bool = True,
        is_writeable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def finalize_configuration(self, model_class: type, name: str) -> None:
        """
        Finalize and check the configuration.

        This is an external trigger called by the model class when the model class is ready.
        The reason it exists here instead of in the constructor is because some columns are tightly
        connected to the model class, and can't validate configuration until they know what the model is.
        Therefore, we need the model involved, and the only way for a property to know what class it is
        in is if the parent class checks in (which is what happens here).
        """
        self.model_class = model_class
        self.name = name
        getattr(self.__class__, "pivot_table_name").finalize_and_validate_configuration(self)
        own_column_name_in_pivot_config = getattr(self.__class__, "own_column_name_in_pivot")
        own_column_name_in_pivot_config.source_model_class = model_class
        own_column_name_in_pivot_config.finalize_and_validate_configuration(self)
        self.finalize_and_validate_configuration()

    def to_backend(self, data):
        # we can't persist our mapping data to the database directly, so remove anything here
        # and take care of things in post_save
        if self.name in data:
            del data[self.name]
        return data

    @property
    def pivot_model(self) -> Model:
        return self.di.build(self.pivot_model_class, cache=True)

    @property
    def related_model(self) -> Model:
        return self.di.build(self.related_model_class, cache=True)

    @property
    def related_columns(self) -> dict[str, Column]:
        return self.related_model.get_columns()

    @property
    def pivot_columns(self) -> dict[str, Column]:
        return self.pivot_model.get_columns()

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> list[str | int]:
        pass

    def __get__(self, instance, cls):
        if instance is None:
            self.model_class = cls
            return self

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            instance.get_columns()

        related_id_column_name = self.related_model_class.id_column_name
        return [getattr(model, related_id_column_name) for model in self.get_related_models(instance)]

    def __set__(self, instance, value: list[str | int]) -> None:
        instance._next_data[self.name] = value

    def get_related_models(self, model: Model) -> Model:
        related_column_name_in_pivot = self.related_column_name_in_pivot
        own_column_name_in_pivot = self.own_column_name_in_pivot
        pivot_table_name = self.pivot_table_name
        related_id_column_name = self.related_model_class.id_column_name
        model_id = getattr(model, self.model_class.id_column_name)
        model = self.related_model
        join = f"JOIN {pivot_table_name} ON {pivot_table_name}.{related_column_name_in_pivot}={model.destination_name()}.{related_id_column_name}"
        related_models = model.join(join).where(f"{pivot_table_name}.{own_column_name_in_pivot}={model_id}")
        return related_models

    def get_pivot_models(self, model: Model) -> Model:
        return self.pivot_model.where(
            f"{self.own_column_name_in_pivot}=" + getattr(model, self.model_class.id_column_name)
        )

    def post_save(self, data: dict[str, Any], model: clearskies.model.Model, id: int | str) -> None:
        # if our incoming data is not in the data array or is None, then nothing has been set and we do not want
        # to make any changes
        if self.name not in data or data[self.name] is None:
            return

        # figure out what ids need to be created or deleted from the pivot table.
        if not model:
            old_ids = set()
        else:
            old_ids = set(self.__get__(model, model.__class__))

        new_ids = set(data[self.name])
        to_delete = old_ids - new_ids
        to_create = new_ids - old_ids
        pivot_model = self.pivot_model
        related_column_name_in_pivot = self.related_column_name_in_pivot
        if to_delete:
            for model_to_delete in pivot_model.where(
                f"{related_column_name_in_pivot} IN ({','.join(map(str, to_delete))})"
            ):
                model_to_delete.delete()
        if to_create:
            own_column_name_in_pivot = self.own_column_name_in_pivot
            for id_to_create in to_create:
                pivot_model.create(
                    {
                        related_column_name_in_pivot: id_to_create,
                        own_column_name_in_pivot: id,
                    }
                )

        super().post_save(data, model, id)

    def add_search(self, model: Model, value: str, operator: str = "", relationship_reference: str = "") -> Model:
        related_column_name_in_pivot = self.related_column_name_in_pivot
        own_column_name_in_pivot = self.own_column_name_in_pivot
        own_id_column_name = self.model_class.id_column_name
        pivot_table_name = self.pivot_table_name
        my_table_name = self.model_class.destination_name()
        related_table_name = self.related_model.destination_name()
        join_pivot = f"JOIN {pivot_table_name} ON {pivot_table_name}.{own_column_name_in_pivot}={my_table_name}.{own_id_column_name}"
        # no reason we can't support searching by both an id or a list of ids
        values = value if type(value) == list else [value]
        search = " IN (" + ", ".join([str(val) for val in value]) + ")"
        return model.join(join_pivot).where(f"{pivot_table_name}.{related_column_name_in_pivot}{search}")

    def to_json(self, model: Model) -> dict[str, Any]:
        related_id_column_name = self.related_model_class.id_column_name
        records = [getattr(related, related_id_column_name) for related in self.get_related_models(model)]
        return {self.name: records}

    def documentation(self, name: str | None = None, example: str | None = None, value: str | None = None):
        related_id_column_name = self.related_model_class.id_column_name
        return AutoDocArray(name if name is not None else self.name, AutoDocString(related_id_column_name))



from typing import Any

from clearskies.di.injectable import Injectable


class ByClass(Injectable):
    def __init__(self, cls: type, cache: bool = True):
        if not isinstance(cls, type):
            raise TypeError(
                f"I expected a class for the first argument to clearskies.di.inject.ByClass, but I received an object of type '{cls.__class__.__name__}' instead."
            )
        self.cls = cls
        self.cache = cache

    def __get__(self, instance, parent) -> Any:
        if instance is None:
            return self  # type: ignore

        if self.cls in self._di._class_overrides_by_class:
            return self._di.build_class(self._di._class_overrides_by_class[self.cls], cache=self.cache)
        return self._di.build_class(self.cls, cache=self.cache)



import requests

from clearskies.di.injectable import Injectable
from clearskies.environment import Environment as EnvironmentDependency


class Environment(Injectable):
    def __init__(self, cache: bool = True):
        self.cache = cache

    def __get__(self, instance, parent) -> EnvironmentDependency:
        if instance is None:
            return self  # type: ignore
        return self._di.build_from_name("environment", cache=self.cache)
